/*
 * pTestWebServer.java
 *
 * Copyright(c) 2001, Particle Corp.
 */

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * Class to test pThreadPool
 *
 * Super simple multithreaded (threadpool) Web Server.
 * Amazingly fast and efficient (considering that it does
 * no caching...)
 *
 * @author Alex S.
 * @version v0.0.1
 */
public class pTestWebServer implements pThreadPoolWorkListener {

    // dir where our http root is (where we have html files)
    public static final String wwwdir = "public_html";

    // need a list of content types...
    private static Hashtable contentTypes = new Hashtable();

    // fill up the list with mime types for common extensions
    static {
        String[][] tmp = {
            {"application/msword","doc"},
            {"application/octet-stream","exe"},
            {"application/octet-stream","class"},
            {"application/pdf","pdf"},
            {"application/x-tar","tar"},
            {"application/zip","zip"},
            {"audio/midi","mid"},
            {"audio/midi","midi"},
            {"audio/mpeg","mp3"},
            {"audio/x-wav","wav"},
            {"image/bmp","bmp"},
            {"image/gif","gif"},
            {"image/jpeg","jpeg"},
            {"image/jpeg","jpg"},
            {"image/png","png"},
            {"image/tiff","tiff"},
            {"image/tiff","tif"},
            {"text/css","css"},
            {"text/html","html"},
            {"text/html","htm"},
            {"text/plain","asc"},
            {"text/plain","txt"},
            {"text/plain","java"},
            {"text/xml","xml"},
            {"video/mpeg","mpeg"},
            {"video/mpeg","mpg"}};
            for(int i=0;i<tmp.length;i++)
                contentTypes.put("."+tmp[i][1],tmp[i][0]);
    }

    /**
     * given a filename, returns the content type for it
     * (uses the contentTypes hashtable built above)
     *
     * @param resource Full resource name (ie: a filename)
     * @return a MIME type corresponding to the resource name
     */
    private String getContentType(String resource){
        int extind = resource.lastIndexOf('.');
        if (extind >= 0) {
            String s = (String)contentTypes.get(resource.substring(extind,resource.length()));
            if (s != null)
                return s;
        }
        return "application/octet-stream";
    }

    /**
     * method to parse out the request and place things into a Hashtable
     *
     * @param stream where to get data from...
     * @return request parameters in hashtable (properties)
     */
    private Properties parseRequestParams(InputStream is) throws IOException {
        Properties hash = new Properties();
        BufferedReader br = new BufferedReader(new InputStreamReader(is));
        StringTokenizer st = new StringTokenizer(br.readLine());
        hash.setProperty("method",st.nextToken().toLowerCase());
        if(st.hasMoreTokens())
            hash.setProperty("resource",st.nextToken());
        if(st.hasMoreTokens())
            hash.setProperty("httpversion",st.nextToken().toLowerCase());
        String line = br.readLine();
        while(line != null && line.length() > 0){
            int ind = line.indexOf(':');
            int len = line.length();
            if(ind > 0 && (ind+1) < len)
                hash.setProperty(line.substring(0,ind).trim().toLowerCase(),
                                 line.substring(ind+1,len).trim());
            line = br.readLine();
        }
        return hash;
    }

    /**
     * write out an error message
     *
     * @param ps PrintStream; where to write
     * @param code The HTTP defined error code...
     * @param message An http message definition of the code...
     */
    public void writeErrorResponse(PrintStream ps,String code,String message){
        ps.println("HTTP/1.1 "+code+" OK\n"+
                   "Content-Type: text/html\n\n"+
                   "<html><head><title>"+code+" "+message+"</title></head>"+
                   "<body><h1>"+code+" "+message+"</h1></body></html>");
    }

    /**
     * produce response (note: this method is kinda similar to the
     * servlet service method... :-)
     *
     * @param request Properties object containing request parameters
     * @param res OutputStream where to write the response
     */
    private void produceResponse(Properties request,OutputStream res) throws IOException {
        PrintStream ps = new PrintStream(res);
        if (((String)request.get("method")).equals("get")) {
            String resource = wwwdir + (String)request.get("resource");
            if (resource.endsWith("/"))
                resource = resource + "index.html"; // only do index.html (no dir listing)
            File file = new File(resource);
            if(!file.exists())
                writeErrorResponse(ps,"404","Not Found");
            else if(!file.canRead())
                writeErrorResponse(ps,"403","Forbidden");
            else {
                ps.println("HTTP/1.1 200 OK");
                ps.println("Content-Type: "+getContentType(resource));
                ps.println("Content-Length: "+file.length()+"\n");
                // write this file to output...
                FileInputStream fis = new FileInputStream(file);
                byte[] buffer = new byte[4096]; // 4k buffers
                int read = fis.read(buffer);
                while(read >= 0){
                    ps.write(buffer,0,read);
                    read = fis.read(buffer);
                }
            }
        } else  // only support GET requests for now...
            writeErrorResponse(ps,"400","Bad Request");
        ps.flush();
    }

    /**
     * workhorse function (part of pThreadPoolWorkListener interface)
     *
     * @param o The connected socket with which to work with.
     */
    public void performWork(Object o){
        Socket socket = (Socket)o;
        try {
            Properties request = parseRequestParams(socket.getInputStream());
            produceResponse(request,socket.getOutputStream());
        } catch (Throwable e){  // mostly for IOException but...
            // sometimes a NullPointerException might show up because browser
            // closed connection without sending initial request header
            // (this does happen [IE])
            // e.printStackTrace();
        } finally {
            try {
                socket.close();
            } catch (Throwable e) {
                // e.printStackTrace();
            }
        }
    }

    /**
     * start the server
     */
    public static void main(String[] args) throws Throwable {
        pThreadPool pool = new pThreadPool(50,new pTestWebServer());
        ServerSocket socket = new ServerSocket(8080,100);
        System.out.println("pTestWebServer: [Copyright(c) 2001, Particle Corp.]\n"+
                           "50 Threads in Thread Pool.\n"+
                           "Listening on port 8080.\n"+
                           "WWW Root Dir is: "+(new File(wwwdir)).getAbsolutePath());
        for (;;)    // loop until someone kills us :-)
            pool.insert(socket.accept());
    }
}

