/*
 * pThreadPool.java
 *
 * Copyright(c) 2001, Particle Corp.
 */

/**
 * General purpose Thread Pool.
 *
 * @author Alex S.
 * @version v0.0.1
 */
public class pThreadPool {

    // queue to hold work data [Note: we do not want to use Vector, it sucks as a queue]
    private pQueue queue = new pQueue();

    // our performer...
    private pThreadPoolWorkListener workerObject = null;

    /**
     * constructor to initialize the pool to a certain number of threads
     *
     * @param threads Number of threads in the pool.
     */
    public pThreadPool(int threads,pThreadPoolWorkListener w){
        if(threads < 1 || w == null)
            throw new java.lang.IllegalArgumentException();
        workerObject = w;
        // create some of our threads and throw them into the pool :-)
        for(int i=0;i<threads;i++){
            pThread thread = new pThread(this);
            thread.start(); // start them off waiting...
        }
    }

    /**
     * Get the work object for processing... synchronized makes sure that
     * only 1 thread executes this at a time.
     *
     * @return A work object ready for processing by one of the threads in the pool.
     */
    private synchronized Object remove(){
        // wait for object to appear
        while(queue.isEmpty()){
            try{
                wait(); // wait until notified
            }catch(InterruptedException ie){}
        }
        // by now, we should be the only thread running this code
        // and the only thread removing an object from a non-empty queue.
        return queue.remove();
    }

    /**
     * inserts an object to be handled by the thread pool
     */
    public synchronized void insert(Object o){
        queue.insert(o);
        notifyAll();    // notify waiting thread
    }

    /**
     * thread pool thread. gets work from thread pool
     * controller, processes it, and asks for more.
     *
     * @author Alex S., Copyright(c) 2000-2001, Particle
     */
    static class pThread extends Thread {

        // link back to thread pool controller
        static pThreadPool controller;

        /**
         * initializes this thread to be part of a certain pool
         *
         * @param threadPool Thread Pool this thread should be part of
         */
        public pThread(pThreadPool threadPool) {
            controller = threadPool;
        }

        /**
         * Thread run method.
         * This is where processing (and waiting for work) happens.
         */
        public void run(){
            for(;;){
                controller.workerObject.performWork(controller.remove());
            }
        }
    }

    /**
     * An array based synchronized Queue, grows if needed.
     * Really fast insert and remove operations.
     * [a LOT faster than using Vector as a queue]
     *
     * @author Alex S., Copyright(c) 1998-2001, Particle
     */
    static class pQueue {

        // array that holds our data
        private Object[] array;

        // start is where things are inserted; end is where they're removed.
        private int start,end;

        /**
         * constructs a new queue
         */
        public pQueue(){
            array = new Object[1024];
            start = end = 0;
        }

        /**
         * is this queue empty?
         *
         * @return true if queue is empty; false otherwise
         */
        public synchronized boolean isEmpty(){
            return start == end;
        }

        /**
         * increases the capacity of this queue by some amount
         *
         * unsynchronized because only called from synchronized
         * methods... this avoids the unnessasary synchronization.
         */
        private void increaseCapacity(){
            Object[] array2 = new Object[array.length + 1024];
            for(int i=0;i<array.length;i++)
                array2[i] = array[end = (++end % array.length)];
            start = array.length - 1;
            array = array2;
            end = array.length - 1;
        }

        /**
         * insert method...
         *
         * @param o The object to insert.
         */
        public synchronized void insert(Object o){
            array[start = (++start % array.length)] = o;
            if(start == end)
                increaseCapacity();
        }

        /**
         * remove method; returns null if queue is empty
         *
         * @return The removed object
         */
        public synchronized Object remove(){
            if(end == start)
                return null;
            return array[end = (++end % array.length)];
        }

        /**
         * print out the objects in this queue, in remove order
         * [this is mostly for debug purposes only]
         *
         * @return String representation of the queue
         */
        public synchronized String toString(){
            StringBuffer sb = new StringBuffer("pQueue["+array.length+"](");
            if(!isEmpty()){
                int i = end;
                do{
                    i = ++i % array.length;
                    sb.append("["+array[i]+"]");
                }while(i != start);
            }
            sb.append(")");
            return sb.toString();
        }
    }
}


